<!doctype html>
<html>
<head>
{% block head %}
<title>{% block title %}{% endblock %}</title>
<link rel=stylesheet type=text/css href="{{ url_for('static', filename='style.css') }}">
{% endblock %}
</head>
<body>
<div class="bar1">
{% block bar1 %}
{% endblock %}
</div>
<div class="bar">
{{ home|safe }}
<a href="https://wiki.fysik.dtu.dk/ase/">ASE</a> ::
<a href="https://wiki.fysik.dtu.dk/ase/ase/db/db.html">ase.db</a> ::
<a href="http://www.python.org/">Python</a> ::
<a href="http://docs.scipy.org/doc/numpy/reference/">NumPy</a>
</div>
<br>
{% block content %}
{% endblock %}
<br>
<div class="bar">
{{ home|safe }}
<a href="https://wiki.fysik.dtu.dk/ase/">ASE</a> ::
<a href="https://wiki.fysik.dtu.dk/ase/ase/db/db.html">ase.db</a> ::
<a href="http://www.python.org/">Python</a> ::
<a href="http://docs.scipy.org/doc/numpy/reference/">NumPy</a>
</div>
<div class="footer">
Created using <a href="http://flask.pocoo.org/">Flask</a>
</div>
</body>
</html>
