# alternative structures
# data from CRC Handbook 2004 85th edition
alternative_structures = [
    None,# X
    None,# H
    None,# He
    None,# Li
    None,# Be
    None,# B
    None,# C
    None,# N
    None,# O
    None,# F
    None,# Ne
    None,# Na
    None,# Mg
    None,# Al
    None,# Si
    None,# P
    None,# S
    None,# Cl
    None,# Ar
    None,# K
    None,# Ca
    None,# Sc
    None,# Ti
    None,# V
    None,# Cr
    None,# Mn
    {'symmetry': 'fcc', 'a': 2.9315, 'comment' : 'T>910 C'},# Fe
    None,# Co
    None,# Ni
    None,# Cu
    None,# Zn
    None,# Ga
    None,# Ge
    None,# As
    None,# Se
    None,# Br
    None,# Kr
    None,# Rb
    None,# Sr
    None,# Y
    None,# Zr
    None,# Nb
    None,# Mo
    None,# Tc
    None,# Ru
    None,# Rh
    None,# Pd
    None,# Ag
    None,# Cd
    None,# In
    None,# Sn
    None,# Sb
    None,# Te
    None,# I
    None,# Xe
    None,# Cs
    None,# Ba
    None,# La
    None,# Ce
    None,# Pr
    None,# Nd
    None,# Pm
    None,# Sm
    None,# Eu
    None,# Gd
    None,# Tb
    None,# Dy
    None,# Ho
    None,# Er
    None,# Tm
    None,# Yb
    None,# Lu
    None,# Hf
    None,# Ta
    None,# W
    None,# Re
    None,# Os
    None,# Ir
    None,# Pt
    None,# Au
    None,# Hg
    None,# Tl
    None,# Pb
    None,# Bi
    None,# Po
    None,# At
    None,# Rn
    None,# Fr
    None,# Ra
    None,# Ac
    None,# Th
    None,# Pa
    None,# U
    None,# Np
    None,# Pu
    None,# Am
    None,# Cm
    None,# Bk
    None,# Cf
    None,# Es
    None,# Fm
    None,# Md
    None,# No
    None,# Lr
    ]
