# Copyright (C) 2005 jrk

""" Default pseudopotential paths are defined here
"""

__docformat__ = 'reStructuredText'

defaultpseudopotentials = {'O':'008-O-gpbe-rc_1_3_nlc.uspp',
                           'Pd':'046-Pd-gpe-n-6projectors-floc.uspp',
                           'Ag':'047-Ag-gpe-n-floc.uspp'}

